package org.apache.bcel.generic;

import java.util.Objects;
import org.apache.bcel.classfile.AnnotationEntry;
import java.util.Iterator;
import org.apache.bcel.classfile.Utility;
import org.apache.bcel.classfile.Constant;
import org.apache.bcel.classfile.ConstantPool;
import org.apache.bcel.classfile.ConstantObject;
import java.util.ArrayList;
import java.util.function.Consumer;
import java.util.stream.Stream;
import org.apache.bcel.classfile.Attribute;
import org.apache.bcel.classfile.Annotations;
import org.apache.bcel.classfile.ConstantValue;
import org.apache.bcel.classfile.Field;
import java.util.List;
import org.apache.bcel.util.BCELComparator;

public class FieldGen extends FieldGenOrMethodGen
{
    private static BCELComparator bcelComparator;
    private Object value;
    private List<FieldObserver> observers;
    
    public static BCELComparator getComparator() {
        return FieldGen.bcelComparator;
    }
    
    public static void setComparator(final BCELComparator comparator) {
        FieldGen.bcelComparator = comparator;
    }
    
    public FieldGen(final Field field, final ConstantPoolGen cp) {
        this(field.getAccessFlags(), Type.getType(field.getSignature()), field.getName(), cp);
        final Attribute[] attributes;
        final Attribute[] attrs = attributes = field.getAttributes();
        for (final Attribute attr : attributes) {
            if (attr instanceof ConstantValue) {
                this.setValue(((ConstantValue)attr).getConstantValueIndex());
            }
            else if (attr instanceof Annotations) {
                final Annotations runtimeAnnotations = (Annotations)attr;
                runtimeAnnotations.forEach(element -> this.addAnnotationEntry(new AnnotationEntryGen(element, cp, false)));
            }
            else {
                this.addAttribute(attr);
            }
        }
    }
    
    public FieldGen(final int accessFlags, final Type type, final String name, final ConstantPoolGen cp) {
        super(accessFlags);
        this.setType(type);
        this.setName(name);
        this.setConstantPool(cp);
    }
    
    private void addAnnotationsAsAttribute(final ConstantPoolGen cp) {
        Stream.of(AnnotationEntryGen.getAnnotationAttributes(cp, super.getAnnotationEntries())).forEach(this::addAttribute);
    }
    
    private int addConstant() {
        switch (super.getType().getType()) {
            case 4:
            case 5:
            case 8:
            case 9:
            case 10: {
                return super.getConstantPool().addInteger((int)this.value);
            }
            case 6: {
                return super.getConstantPool().addFloat((float)this.value);
            }
            case 7: {
                return super.getConstantPool().addDouble((double)this.value);
            }
            case 11: {
                return super.getConstantPool().addLong((long)this.value);
            }
            case 14: {
                return super.getConstantPool().addString((String)this.value);
            }
            default: {
                throw new IllegalStateException("Unhandled : " + super.getType().getType());
            }
        }
    }
    
    public void addObserver(final FieldObserver o) {
        if (this.observers == null) {
            this.observers = new ArrayList<FieldObserver>();
        }
        this.observers.add(o);
    }
    
    public void cancelInitValue() {
        this.value = null;
    }
    
    private void checkType(final Type atype) {
        final Type superType = super.getType();
        if (superType == null) {
            throw new ClassGenException("You haven't defined the type of the field yet");
        }
        if (!this.isFinal()) {
            throw new ClassGenException("Only final fields may have an initial value!");
        }
        if (!superType.equals(atype)) {
            throw new ClassGenException("Types are not compatible: " + superType + " vs. " + atype);
        }
    }
    
    public FieldGen copy(final ConstantPoolGen cp) {
        final FieldGen fg = (FieldGen)this.clone();
        fg.setConstantPool(cp);
        return fg;
    }
    
    @Override
    public boolean equals(final Object obj) {
        return FieldGen.bcelComparator.equals(this, obj);
    }
    
    public Field getField() {
        final String signature = this.getSignature();
        final int nameIndex = super.getConstantPool().addUtf8(super.getName());
        final int signatureIndex = super.getConstantPool().addUtf8(signature);
        if (this.value != null) {
            this.checkType(super.getType());
            final int index = this.addConstant();
            this.addAttribute(new ConstantValue(super.getConstantPool().addUtf8("ConstantValue"), 2, index, super.getConstantPool().getConstantPool()));
        }
        this.addAnnotationsAsAttribute(super.getConstantPool());
        return new Field(super.getAccessFlags(), nameIndex, signatureIndex, this.getAttributes(), super.getConstantPool().getConstantPool());
    }
    
    public String getInitValue() {
        if (this.value != null) {
            return this.value.toString();
        }
        return null;
    }
    
    @Override
    public String getSignature() {
        return super.getType().getSignature();
    }
    
    @Override
    public int hashCode() {
        return FieldGen.bcelComparator.hashCode(this);
    }
    
    public void removeObserver(final FieldObserver o) {
        if (this.observers != null) {
            this.observers.remove(o);
        }
    }
    
    public void setInitValue(final boolean b) {
        this.checkType(Type.BOOLEAN);
        if (b) {
            this.value = 1;
        }
    }
    
    public void setInitValue(final byte b) {
        this.checkType(Type.BYTE);
        if (b != 0) {
            this.value = b;
        }
    }
    
    public void setInitValue(final char c) {
        this.checkType(Type.CHAR);
        if (c != '\0') {
            this.value = c;
        }
    }
    
    public void setInitValue(final double d) {
        this.checkType(Type.DOUBLE);
        if (d != 0.0) {
            this.value = d;
        }
    }
    
    public void setInitValue(final float f) {
        this.checkType(Type.FLOAT);
        if (f != 0.0) {
            this.value = f;
        }
    }
    
    public void setInitValue(final int i) {
        this.checkType(Type.INT);
        if (i != 0) {
            this.value = i;
        }
    }
    
    public void setInitValue(final long l) {
        this.checkType(Type.LONG);
        if (l != 0L) {
            this.value = l;
        }
    }
    
    public void setInitValue(final short s) {
        this.checkType(Type.SHORT);
        if (s != 0) {
            this.value = s;
        }
    }
    
    public void setInitValue(final String str) {
        this.checkType(ObjectType.getInstance("java.lang.String"));
        if (str != null) {
            this.value = str;
        }
    }
    
    private void setValue(final int index) {
        final ConstantPool cp = super.getConstantPool().getConstantPool();
        final Constant c = cp.getConstant(index);
        this.value = ((ConstantObject)c).getConstantValue(cp);
    }
    
    @Override
    public final String toString() {
        String access = Utility.accessToString(super.getAccessFlags());
        access = (access.isEmpty() ? "" : (access + " "));
        final String signature = super.getType().toString();
        final String name = this.getName();
        final StringBuilder buf = new StringBuilder(32);
        buf.append(access).append(signature).append(" ").append(name);
        final String value = this.getInitValue();
        if (value != null) {
            buf.append(" = ").append(value);
        }
        return buf.toString();
    }
    
    public void update() {
        if (this.observers != null) {
            for (final FieldObserver observer : this.observers) {
                observer.notify(this);
            }
        }
    }
    
    static {
        FieldGen.bcelComparator = new BCELComparator() {
            @Override
            public boolean equals(final Object o1, final Object o2) {
                final FieldGen THIS = (FieldGen)o1;
                final FieldGen THAT = (FieldGen)o2;
                return Objects.equals(THIS.getName(), THAT.getName()) && Objects.equals(THIS.getSignature(), THAT.getSignature());
            }
            
            @Override
            public int hashCode(final Object o) {
                final FieldGen THIS = (FieldGen)o;
                return THIS.getSignature().hashCode() ^ THIS.getName().hashCode();
            }
        };
    }
}
